# encoding: utf-8

import json


class Node(object):


    def __init__(self, parent=None, key=None, value=0, info=None):

        self.reset()

        self.parent   = parent
        self.key      = key
        self.value    = value

        if info is not None:
            self.add(info)



    def reset(self):

        self.setas_root()

        self.key      = None
        self.value    = 0
        self.infoes   = set()

        self.setas_leaf()



    def add(self, info):

        self.infoes.add(info)



    def delete(self, info):

        self.infoes = self.infoes - {info}



    @property
    def empty(self):

        return not bool(self.infoes)



    def setas_root(self):

        self.parent   = None



    def setas_leaf(self):

        self.children = {}



    def root(self):

        return self.parent is None



    def leaf(self):

        return not bool(self.children)



    def meaningless(self):

        return 0 == self.value



    def __repr__(self):

        return ', '.join((str(id(self.parent)),
                         str(self.key),
                         str(self.value),
                         str(self.root()),
                         str(self.leaf()),
                         json.dumps({
                             k:id(v) for k, v in self.children.items()
                         }, ensure_ascii=False)))





class NodeBinary(Node):



    def setas_leaf(self):

        self.children = []


    @property
    def child_left(self):

        try:
            return self.children[0]
        except:
            return None



    @child_left.setter
    def child_left(self, node):

        try:
            self.children[0] = node
        except:
            self.children.append(node)



    @property
    def child_right(self):

        try:
            return self.children[1]
        except:
            return None



    @child_right.setter
    def child_right(self, node):

        try:
            self.children[1] = node
        except:
            self.children.append(node)



    def left_only(self):

        return 1 == len(self.children)



    def __repr__(self):

        return ', '.join((str(id(self.parent)),
                          str(id(self)),
                          str(self.key),
                          str(self.value),
                          str(self.root()),
                          str(self.leaf()),
                          json.dumps([id(v) for v in self.children], ensure_ascii=False),
                          json.dumps([id(v) for v in self.infoes]  , ensure_ascii=False),
                          str(self.empty)))



# encoding: utf-8

import numpy   as np
import _pickle as pickle

from .node     import NodeBinary
from .sequence import Sequence


class Heap(object):


    def __init__(self):

        self.reset()



    def reset(self):

        self._heap = None
        self._seq  = Sequence()

        self.value_sum = 0



    @property
    def empty(self):

        return self._heap is None



    @property
    def node_max(self):

        pass



    @property
    def node_min(self):

        pass



    def swap(self):

        pass



    def tune_up(self, node):

        pass



    def tune_down(self, node):

        pass



    def tune_init(self):

        for i in range(self._seq.tail_index(),-1,-1):
            self.tune_down(self._seq[i])



    def tune(self, node):

        if node.root():
            self.tune_down(node)

        elif node.leaf():
            self.tune_up(node)



    def add(self, value):

        pass



    def delete(self):

        pass



    def save(self, fn):

        with open(fn, 'wb') as f:
            pickle.dump(self.__dict__, f)



    def load(self, fn):

        with open(fn, 'rb') as f:
            d = pickle.load(f)
            self.__dict__.update(d)





class HeapPair(Heap):



    def tune_down(self, node):

        nodes = [node]

        while(nodes):

            n = nodes.pop()
            nodes.extend(self.swap_down(n))



    def tune_up(self, node):

        nodes = [node]

        while(nodes):

            n = nodes.pop()
            nodes.extend(self.swap_up(n))
         


    def add(self, key, value, info):

        # TODO: make node --> add_bynode

        # update value sum
        self.value_sum += key * value

        if self._heap is None:
            node       = NodeBinary(key=key, value=value, info=info)
            self._heap = node
            self._seq.append(self._heap)
        else:
            # new node
            node_parent = self._seq.tail()
            node        = NodeBinary(parent=node_parent, key=key, value=value, info=info)

            # refresh node parent
            node_parent.children.append(node)

            # refresh heap
            self._seq.append(node)

            # tune
            self.tune_up(node)

        return node



    def add_bynode(self, node):

        # update value sum
        self.value_sum += node.key * node.value

        if self._heap is None:
            node.setas_root()
            self._heap = node
            self._seq.append(self._heap)
        else:
            node_parent = self._seq.tail()

            node.setas_leaf()
            node.parent = node_parent

            # refresh node parent
            node_parent.children.append(node)

            # refresh heap
            self._seq.append(node)

            # tune
            self.tune_up(node)

        return node




    def delete(self, key):

        node = self[key]

        self.delete_bynode(node)



    def delete_bynode(self, node):

        if 1 == len(self._seq) and node.root():
            self.reset()

        else:
            # update value sum
            self.value_sum -= node.key * node.value

            node_last = self._seq.last()

            if node == node_last:
                # update seq
                self._seq.last_drop()

                # update descendant info
                node.parent.children = [n for n in node.parent.children if n != node]

            else:
                # update seq
                idx = self._seq.index(node)
                self._seq[idx] = node_last
                self._seq.last_drop()

                # update descendant info
                node_last.parent.children = [n for n in node_last.parent.children
                                               if n != node_last and n != node]

                # update tree
                node_last.parent   = node.parent
                node_last.children = [n for n in node.children if n != node_last]

                # update ancestor and descendant info
                if node.parent:
                    if node.parent.child_left == node:
                        node.parent.child_left = node_last
                    else:
                        node.parent.child_right = node_last

                for n in node.children:
                    n.parent = node_last

                # update root
                if node_last.root():
                    self._heap = node_last

                # tune
                self.tune(node_last)




    def delete_byinfo(self, key, value, info):

        # update tree and seq
        node = self[key]

        if node:
            node.value -= value
            node.delete(info)
            self.tune(node)

            # update value sum
            self.value_sum -= key * value



    def __getitem__(self, key):

        for node in self._seq:
            if key == node.key:
                return node

        return None



    def update(self, key, value):

        # TODO: merge with update_byinfo

        # update value sum
        self.value_sum += key * value

        # update tree and seq
        node = self[key]

        if node:
            node.value += value
            self.tune(node)
        else:
            self.add(key, value)



    def update_byinfo(self, key, value, info):

        # update value sum
        self.value_sum += key * value

        # update tree and seq
        node = self[key]

        if node:
            node.value += value
            node.add(info)
            self.tune(node)
        else:
            node = self.add(key, value, info)

        return node



# encoding: utf-8

import numpy as np

from .heap import Heap
from .heap import HeapPair


class HeapBrother(Heap):


    pass




class HeapBrotherPair(HeapBrother, HeapPair):


    pass




class HeapBrotherPairMax(HeapBrotherPair):



    @property
    def node_max(self):

        return self._heap




    @property
    def node_min(self):
        if self.empty:
            return None

        if 1 == len(self._seq):
            return self._heap

        idx = self._seq.tail_index()

        node = self._seq.tail()
        for n in self._seq[idx+1:]:
            if n.key < node.key:
                node = n

        return node




    def swap(self, node):

        # update tree
        ancestor         = node.parent
        descendant_left  = node.child_left.children
        descendant_right = node.child_right.children

        nodes      = np.array([node, node.child_left, node.child_right])
        nodes_keys = [n.key for n in nodes]

        idx = np.argsort(nodes_keys)[::-1]
        
        root, left, right = nodes[idx]

        root.parent = ancestor
        root.setas_leaf()
        root.children.append(left)
        root.children.append(right)

        left.parent    = root
        left.children  = descendant_left
        
        right.parent   = root
        right.children = descendant_right

        # update ancestor and descendant info
        if ancestor:
            if ancestor.child_left == node:
                ancestor.child_left = root
            else:
                ancestor.child_right = root

        for n in descendant_left:
            n.parent = left

        for n in descendant_right:
            n.parent = right

        # update seq
        nodes_index = [self._seq.index(n) for n in nodes]
        nodes       = nodes[idx]
        self._seq[nodes_index] = nodes

        # update root
        if root.root():
            self._heap = root

        return root, idx



    def swap_down(self, node):

        if node.leaf():
            return []

        elif node.left_only():

            if node.key >= node.child_left.key:
                return []
            else:
                # update root
                if node.root():
                    self._heap = node.child_left

                # update seq
                nodes       = [node, node.child_left]
                nodes_index = [self._seq.index(n) for n in nodes]
                nodes       = nodes[::-1]
                self._seq[nodes_index] = nodes

                # update tree
                ancestor   = node.parent
                descendant = node.child_left.children

                # update ancestor and descendant info
                if ancestor:
                    if ancestor.child_left == node:
                        ancestor.child_left = node.child_left
                    else:
                        ancestor.child_right = node.child_left

                for n in descendant:
                    n.parent = node

                # update tree
                node.child_left.parent = ancestor
                node.child_left.setas_leaf() 
                node.child_left.children.append(node)

                node.parent   = node.child_left
                node.children = descendant

                return [node]

        else:
            root, idx = self.swap(node)

            tbd = []
            if 1 != idx[1]:
                tbd.append(root.child_left)
            if 2 != idx[2]:
                tbd.append(root.child_right)

            return tbd



    def swap_up(self, node):

        if not node or node.root():
            return []

        elif node.parent.left_only():

            if node.parent.key >= node.key:
                return []
            else:
                node_parent = node.parent

                # update root
                if node_parent.root():
                    self._heap = node

                # update seq
                nodes       = [node_parent, node]
                nodes_index = [self._seq.index(n) for n in nodes]
                nodes       = nodes[::-1]
                self._seq[nodes_index] = nodes

                # update tree
                ancestor   = node_parent.parent
                descendant = node.children
                node.setas_root()
                node.setas_leaf()

                # update ancestor and descendant info
                if ancestor:
                    if ancestor.child_left == node_parent:
                        ancestor.child_left = node
                    else:
                        ancestor.child_right = node

                node_parent.setas_root()

                for n in descendant:
                    n.parent = node_parent

                # update tree
                node_parent.children = descendant

                node.child_left      = node_parent
                node.parent          = ancestor

                node_parent.parent   = node

                return [node.child_left]

        else:
            root, idx = self.swap(node.parent)

            tbd = []
            if 0 != idx[0]:
                tbd.append(root)
            if 1 != idx[1]:
                tbd.append(root.child_left)
            if 2 != idx[2]:
                tbd.append(root.child_right)

            return tbd



    def tune(self, node):

        if node.root():
            self.tune_down(node)

        elif node.leaf():
            self.tune_up(node)

        elif node.child_left and node.child_left.key > node.key:
            self.tune_down(node)

        elif node.child_right and node.child_right.key > node.key:
            self.tune_down(node)

        else:
            self.tune_up(node)



    def have(self, key):

        nodes = [self._heap]

        while(nodes):
            node_current = nodes.pop()

            if key == node_current.key:
                return True

            elif node_current.child_left and key == node_current.child_left.key:
                return True

            elif node_current.child_right and key == node_current.child_right.key:
                return True

            elif node_current.child_right and key < node_current.child_right.key:
                nodes.append(node_current.child_right)
                nodes.append(node_current.child_left)

            elif node_current.child_left and key < node_current.child_left.key:
                nodes.append(node_current.child_left)

        return False



# encoding: utf-8

import math
import numpy as np


class Sequence(object):

    
    def __init__(self, values=[]):

        self._seq  = np.array(values)



    def index(self, value):

        return np.where(self._seq == value)[0][0]



    def tail_index(self):

        return math.floor((len(self._seq)+1)/2) - 1 



    def tail(self):

        return self._seq[self.tail_index()]



    def last(self):

        return self._seq[-1]



    def last_drop(self):

        self._seq = self._seq[:-1]



    def __getitem__(self, key):

        return self._seq[key]



    def __setitem__(self, key, value):

        self._seq[key] = value



    def append(self, value):

        self._seq = np.concatenate((self._seq, [value]))



    def __repr__(self):

        return '\n'.join([v.__repr__() for v in self._seq])


    def __len__(self):

        return len(self._seq)
